#include <stdio.h>
#include <sunmath.h>
#include "scanner.h"

#define BIGASS_STRING 254
#define NUM_CMDS 16
#define NUM_REGS 8
#define LONGEST_CMD 6
#define INST_SIZE 6
#define REG_SIZE 3
#define DATA_SIZE 8
#define MAX_NUM_INSTRUCTIONS 255

static char cmds_ascii[][NUM_CMDS] = { "LOAD", "MOVE", "BRA", "COMP", "INC", "SLL", "SRL", "SRA", "BRZA", "BRZ", "BRN", "ADD", "DEC", "AND", "OR", "LOADIO"};

static char regs_ascii[][NUM_REGS] = { "IRLO", "PC", "FLAGS", "D3", "D4", "D5", "D6", "D7" };

char* get_cmd(int line_num);
char* fmt_cmd(char *cmd);
char* dec2bin(char *num);
char* inst2bin(char *inst);
char* reg2bin(char *reg);
char* pad_inst(char *inst, int to_size);
void print_info(void);
void print_header(char *fn, FILE *of);
void print_footer(char *fn, FILE *of);

int main(int argc, char *argv[]){
	FILE *outfile;
	int i=0, num_exits=0;
	char *buffer;
	char *line_num=(char*)malloc(sizeof(char)*(int)(log2(MAX_NUM_INSTRUCTIONS)+2));

	if(argc > 1 && strchr(argv[1], 'z') != NULL){print_info(); exit(0);}
	printf("%s:\ta tiny assembler to create ROM images for cs23:\n", argv[0]);
	printf("bugs can be mailed to: finney@cs.swarthmore.edu\n");
	printf("for a list of commands and registers run %s -z\n", argv[0]);

	outfile=fopen("default.mif", "w");
	print_header("header", outfile);

	for(i=0;(i<MAX_NUM_INSTRUCTIONS && num_exits!=2); i++){
		buffer=(char*)get_cmd(i);
		buffer=fmt_cmd(buffer);
		if(strcmp(buffer, "exit") == 0){
			i--;
			num_exits++;
		}
		else if(strstr(buffer, "error") != NULL){
			fprintf(stderr, "syntax error, please try again\n");
			i--;
			num_exits=0;
		}
		else{
			sprintf(line_num, "%d", i);	
			line_num=pad_inst(dec2bin(line_num), (int)log2(MAX_NUM_INSTRUCTIONS));
			fprintf(outfile, "\t%s: %s;\n", line_num, buffer);
			num_exits=0;
		}
		if(num_exits==1) fprintf(stderr, "type one more blank line to quit\n");
	}
	print_footer("footer", outfile);
	printf("ROM written to default.mif\n");
	exit(0);
}

char* get_cmd(int line_num){
	char*foo=(char*)malloc(sizeof(char)*BIGASS_STRING+1);
	printf("%d:", line_num);
	fgets(foo, BIGASS_STRING, stdin);
	foo[strlen(foo)-1] = '\0';  //bye bye newline
	return foo;
}

char* fmt_cmd(char *cmd){
	int i, j, found=0;
	char *inst;
	char *reg1;
	char *reg2;
	char *data_or_address;
	char *final;

	InitScanner(cmd);
	IgnoreSpace(1);

	if(AtEndOfLine()) return "exit";
	
	inst=GetNextToken();
	for(i=0;(i<strlen(inst)&&!iswspace(inst[i]));i++){
		if(isalpha(inst[i])) inst[i]=toupper(inst[i]);
		else{fprintf(stderr, "illegal instruction.\n"); return "error";}
	}
	inst=inst2bin(inst);
	if(strcmp(inst, "exit") == 0) return "exit";
	inst=pad_inst(inst, INST_SIZE);

	if(AtEndOfLine()) return "error";
	reg1=GetNextToken();
	reg1=reg2bin(reg1);

	if(AtEndOfLine()) return "error";
	reg2=GetNextToken();
	reg2=reg2bin(reg2);

	if(AtEndOfLine()) return "error";
	data_or_address=GetNextToken();
	data_or_address=pad_inst(dec2bin(data_or_address), DATA_SIZE);

	final=(char*)malloc(sizeof(char)*(strlen(inst)+strlen(reg1)+strlen(reg2)+strlen(data_or_address)+1));
	sprintf(final, "%s%s%s%s", inst, reg1, reg2, data_or_address);
	return final;
}

char* dec2bin(char *str_num){
	int i, j, num;
	char*done;
	for(i=0;i<strlen(str_num);i++){
		if(!isdigit(str_num[i])){
			perror("illegal value passed to dec2bin"); exit(-1);
		}
	}num=atoi(str_num);

	if(num==0) j=0;
	else j=log2(num);
	done=(char*)malloc(sizeof(char)*(j+2));
  done[j+1]='\0';
	for(i=num; i>0; i/=2){
		if(i%2 != 0) done[j--]='1';
		else done[j--]='0';
	}
	if(num==0) done[0]='0';
	return done;
}

char* inst2bin(char *inst){
	int i, found=0;
	for(i=0;i<NUM_CMDS;i++){
		if(strlen(inst) != strlen(cmds_ascii[i])) continue;
		else if(strncmp(inst, cmds_ascii[i], strlen(inst)) == 0){
			found=1;
			break;
		}
	} 
	sprintf(inst, "%d\0", i);
	if(found)return dec2bin(inst);
	else return "error";
}

// pad_inst
// takes a "binary" string and pads it to to_size with leading 0's
char* pad_inst(char *inst, int to_size){
	int i, num_to_pad;
	char* padded_val=(char*)malloc(sizeof(char)*(to_size+1));
	padded_val[to_size]='\0';
	num_to_pad = to_size-strlen(inst);

	for(i=0;i<num_to_pad;i++) padded_val[i]='0';
	for(;i<to_size;i++) padded_val[i]=inst[to_size-i-1];
	return padded_val;
}

// print_header
// prints the header template of the .mif file
void print_header(char *fn, FILE *of){
	FILE *hf=fopen(fn, "r");
	char hbuff[BIGASS_STRING];
	while(fgets(hbuff, BIGASS_STRING, hf)) fprintf(of, "%s", hbuff);

}

// print_footer
// prints the footer template of the .mif file
void print_footer(char *fn, FILE *of){
	FILE *ff=fopen(fn, "r");
	char fbuff[BIGASS_STRING];
	while(fgets(fbuff, BIGASS_STRING, ff)) fprintf(of, "%s", fbuff);
}


// reg2bin
// converts a register name to it's "binary" representation
char* reg2bin(char *reg){
	int i, dumb=NUM_REGS;
	char *reg_num;
	fflush(stdout);

	for(i=0;dumb>=10;i++) dumb /= 10;
	reg_num=(char*)malloc(sizeof(char)*(i+2));

	for(i=0;i<strlen(reg);i++){
		if(isalpha(reg[i])) reg[i]=toupper(reg[i]);
		}

	for(i=0;i<NUM_REGS;i++){
		if(strcmp(reg, regs_ascii[i]) == 0){
			sprintf(reg_num, "%d", i);
			return pad_inst(dec2bin(reg_num), REG_SIZE);
		}
	}

	fprintf(stderr, "invalid register.\n");
	return "error";
}

void print_info(void){
	int i;
	printf("valid commands are:\n");
	for(i=0;i<NUM_CMDS;i++) printf("\t%s", cmds_ascii[i]);
	printf("\nvalid registers are:\n");
	for(i=0;i<NUM_REGS;i++) printf("\t%s", regs_ascii[i]);
	printf("\n");
}
